#!/usr/bin/env python

"""
Take 2 sets of raster images and combine those of the same resolution.


Copyright 2010, 2011 Thorsten Wilms
License: GPLv3, see docs/COPYING

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
from __future__ import division
from os import listdir
from os.path import isdir, isfile
from sys import stderr, stdout
import imghdr
import Image

# Custom modules
import backtestground.common as common
from backtestground.alpha_composite import alpha_composite
from backtestground.common import filename_without_ext
from backtestground.common import join_make_sure_exists
from backtestground.common import progress_n_of_m


def get_image_paths(path):
    """List of single image if path specifies an image file,
    or images found in a directory. Else: empty list.

    """
    issue = []

    # Collect paths
    if isfile(path):
        candidates = [path]

        # Message to be used if this is later found to be no image file:
        if_issue = "%s is not an image file." % (path)
    elif isdir(path):
        # Get items in the directory
        items = listdir(path)

        # Make sure paths ends with "/"
        if path[-1] != "/":
            path = ''.join([path, "/"])

        # Prepend path to item names
        candidates = [''.join([path, i]) for i in items]

        # Message to be used if this is later found to contain no image files
        if_issue = "%s does not contain image files." % (path)
    else:
        candidates = ()
        if_issue = "%s is not a valid file or directory." % (path)

    # Filter for paths to images
    image_paths = filter(lambda path: imghdr.what(path) != None, candidates)
    if not image_paths:
        issue = if_issue

    return image_paths, issue


def get_image_resolution(path):
    """Take a path to an image file and return a resolution, path pair."""

    # Open as file, not directly via Image.open, so it can be closed explicitly
    f = open(path, 'r')
    image = Image.open(f)
    size = image.size
    f.close

    return size, path


def to_pairs(path_sets):
    """From sets of valid image paths to a list of matching resolution pairs.

    """
    # Get resolution/path-to-image pairs
    resolution_path_sets = [map(get_image_resolution, path_sets[i])
                            for i in range(len(path_sets))]

    # Sort. Results in rising width order
    for e in resolution_path_sets:
        e.sort()

    # For every element in the first list, walk through the second list.
    # Add the paths of matches by resolution to 'pairs'.
    # As the lists are sorted, reset the index only to the first match,
    # as to not start comparing to smaller resolutions again.
    pairs = []
    r = 0
    first_hit = len(resolution_path_sets[0]) + 1

    for e in resolution_path_sets[0]:
        while r < len(resolution_path_sets[1]):

            # Do the resolutions match?
            if e[0] == resolution_path_sets[1][r][0]:
                #print "eq"
                pair = (e[01], resolution_path_sets[1][r][1])
                pairs.append(pair)
                first_hit = min(first_hit, r)

            # Is the first width larger than the second?
            elif e[0][0] < resolution_path_sets[1][r][0][0]:
                break

            r += 1

        if first_hit > len(resolution_path_sets[0]):
            r = first_hit
        else:
            r = 0

    return pairs


def per_pair(path, names, pair):
    """Work to be done per pair of images of matching resolution,
    including saving the output image.

    """
    # Open as files, not directly via Image.open,
    # so they can be closed explicitly
    files = map(open, pair)
    images = map(Image.open, files)

    # Make sure there is an alpha channel
    images_rgba = map(lambda i: i.convert('RGBA'), images)

    # Compose
    image = alpha_composite(*images_rgba)

    # Close input files
    map(lambda f: f.close(), files)

    # Save output file
    file_names = map(filename_without_ext, pair)
    image.save(''.join([path, "/",
                        file_names[0], "_under_", file_names[1], ".png"]))


def dispatch(sets, destination, overwrite):
    """Job control."""

    # Get paths to image files found at or in the 2 paths given in sets
    # combined with potential issues (error messages)
    paths_issue_lists = map(get_image_paths, sets)
    path_sets, maybe_issues = zip(*paths_issue_lists)

    # remove empty items from list of issues and
    # append "\n" to each remaining item
    issues = filter(lambda issue: issue, maybe_issues)
    issues_n = [''.join([issues[i], "\n"]) for i in range(len(issues))]

    if issues:
        # At least one of the sets does not contain any valid image file,
        # so put out a message and stop.
        map(stderr.write, issues_n)
    else:
        # Each set contains at least one image file to work with
        pairs = to_pairs(path_sets)

        # Get the filename without extension for single file sets,
        # or the name of the last directory.
        names = []
        for e in sets:
            if isfile(e):
                names.append(filename_without_ext(e))
            else:
                splitted = e.split("/")
                if splitted[-1]:
                    dirname = splitted[-1]
                else:
                    dirname = splitted[-2]

                names.append(dirname)

        # Make sure the output dir exists, move existing to backup,
        # depending on overwrite
        path = join_make_sure_exists([destination, "/", names[0], "_under_",
                                     names[1]], overwrite)

        # Do the rest of work per pair
        total = len(pairs)
        for i in range(total):
            per_pair(path, names, pairs[i])
            # Indicate progress
            progress_n_of_m(total, i)

        stdout.write("Saved results in %s.\n" % path)
