#!/usr/bin/env python

"""
Functions used in both bgc-sizes and sizes.


Copyright 2011 Thorsten Wilms
License: GPLv3, see docs/COPYING

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
from __future__ import division
import os
from string import rfind
from sys import stdout
import Image

# Custom modules
import backtestground.resolutions_conf as resolutions_conf


def filename_without_ext(name):
    """Get the filename without extension."""

    basename = os.path.basename(name)
    last_dot = rfind(basename, ".")

    return basename[:last_dot]


def move_to_numbered_backup(path):
    """Move path to path appended with _old_n, where n is the first number that
    makes the addition not match an existing item.

    """
    # path_old is actually a new path, but contains the term "old"
    path_old = ''.join([path, ".old_"])

    i = 1
    while os.path.exists(''.join([path_old, str(i)])):
        i += 1

    path_old = ''.join([path_old, str(i)])
    os.rename(path, path_old)

    path_rel = os.path.relpath(path)
    path_old_rel = os.path.relpath(path_old)
    text = "Moved existing %s to %s.\n" % (path_rel, path_old_rel)
    stdout.write(text)
    stdout.write("Use -f if you don't want this behavior.\n")
    stdout.flush()


def join_make_sure_exists(path_parts, overwrite):
    """Join path_parts and create directory, if it doesn't exist already."""

    path = ''.join(path_parts)
    if os.path.isdir(path):
        if not overwrite:
            move_to_numbered_backup(path)
            os.mkdir(path)
    else:
        os.mkdir(path)

    return path


def input_resolution_shortcut(resolutions, image, path):
    """Remove the resolution of the input image from resolutions."""
    try:
        resolutions.remove(image.size)
    except ValueError:
        # Don't care if the resolution is not in the list, except for setting
        # the progress indicator offset to zero
        i_offset = 0
    else:
        # Save input image to output directory,
        # as it matches one of the resolutions

        # Save with filenames like 1024x0768.png
        # (zero padded on the left to 4 digits)
        image.save(''.join([path, "/",
                            str(image.size[0]).zfill(4), "x",
                            str(image.size[1]).zfill(4),
                            ".png"]))

        # Print first step of a progress counter
        progress_n_of_m(len(resolutions), 0)
        i_offset = 1

    return resolutions, i_offset


def progress_n_of_m(total, i):
    """Show progress with a 'n of m' output to stdout."""

    max_digits = len(str(total))
    of_n = ''.join([" of ", str(total)])
    if i > 0:
        # Enough backspaces to clear the previous output
        prev_digits = len(''.join([str(i - 1).zfill(max_digits), of_n]))
        for n in range(prev_digits):
            stdout.write("\b")
    stdout.write(''.join([str(i + 1).zfill(max_digits), of_n]))

    if i + 1 == total:
        # Make sure the prompt will appear below,
        # not to the right of the last output
        stdout.write("\n")
    stdout.flush()


def once_for_all_resolutions(input, function_per_res, destination, overwrite):
    """Crop/resize an image that might contain panels, windows and icons to
    several common screen resolutions and take care of content.

    Keyword arguments:
    input -- a function evaluated to a list and a file_name.
             The list contains just an image or an image followed by regions.
    function_per_res -- a function to be called for per resolution work.
    destination -- path to output destination (sans the final directory).

    """
    regions, file_name = input

    # Prepare destination directory, if it doesn't exist already
    path = join_make_sure_exists([destination, "/", file_name, "_resolutions"],
                                 overwrite)

    # Get list of resolutions
    resolutions = resolutions_conf.read_create_or_default()

    # Remove the resolution of the input image from the list and just copy it,
    # if it matches.
    # On success, i_offset will be 1, to make common.progress_n_of_m skip start
    # at 2, when called for each remaining resolution.
    resolutions,  i_offset = input_resolution_shortcut(resolutions,
                                                       regions[0], path)

    # Just use the sole item, if it is just the image,
    # or take the tail of the list, if it is longer, containing regions
    regions_part = regions[min(len(regions) - 1, 1):]

    # Hand over work per resolution
    for i in range(len(resolutions)):
        # material is regions for 'bgc-sizes' and image for 'sizes'.
        function_per_res(regions_part, (resolutions[i][0], resolutions[i][1]),
                         path)

        # Print a progress counter, updated in place
        progress_n_of_m(len(resolutions) + i_offset, i + i_offset)

    stdout.write("Saved results in %s.\n" % path)
