#!/usr/bin/env python

"""
Takes an image with transparent background, that may contain panels, icons and
windows and crop/resizes it to several common screen resolutions, while doing
something hopefully sensible with the content.


Copyright 2010, 2011 Thorsten Wilms
License: GPLv3, see docs/COPYING

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
from __future__ import division
import Image
import ImageChops
from functools import partial
from math import ceil

# Custom modules
import backtestground.common as common
from backtestground.alpha_composite import alpha_composite


def moving_box(direction, length, left_or_upper, right_or_lower):
    """Bounding box start and iteration values to move the box across an image.

    Keyword arguments:
    direction -- string, direction the box will be moved: rightwards,
                 downwards, leftwards, upwards
    length -- length of the path the box will be moved on
    left_or_upper -- coordinate of the upper or left boundary of the box and
                     its path
    right_or_lower -- coordinate of the lower or left boundary of the box and
                      its path

    """
    # With this tuple, the correct value of:
    #   start_right is one step to the right of the value of start_left,
    #   start_bottom is one step to the right of the value of start_upper.
    v = (0, 1, left_or_upper, right_or_lower, length -1, length)

    x = []
    # Set values based on direction
    left_i, upper_i, horizontal, vertical = {
        'rightwards': lambda x: [0, 2,  1,  0],
        'downwards':  lambda x: [2, 0,  0,  1],
        'leftwards':  lambda x: [4, 2, -1,  0],
        'upwards':    lambda x: [2, 4,  0, -1]
    }[direction](x)

    left = v[left_i]
    upper = v[upper_i]
    right = v[left_i + 1]
    lower = v[upper_i + 1]

    return left, upper, right, lower, horizontal, vertical


def moving_box_at_i((left, upper, right, lower, horizontal, vertical), i):
    """Bounding box coordinates during iteration."""

    left_ = left + i * horizontal
    upper_ = upper + i * vertical
    right_ = right + i * horizontal
    lower_ = lower + i * vertical

    return left_, upper_, right_, lower_


def pixel_line(image, moving_box, i):
    """Cut a line from the image."""

    left, upper, right, lower = moving_box_at_i(moving_box, i)
    line = image.crop((left, upper, right, lower))

    return list(line.getdata())


def panel_diameter(image, direction, length, left_or_upper, right_or_lower):
    """Panel (or shadow at the edge) diameter. One pass, applied 4
    times via panel_diamters.

    """
    local_pixel_line = partial(pixel_line, image,
                               moving_box(direction, length, left_or_upper,
                                          right_or_lower))

    for i in range(length):
        line = local_pixel_line(i)

        # Minimum percentage of pixels that must be zero (transparent)
        # for a line to be considered not part of a panel.
        zero_threshold = 90

        # Percentage of list items that are zero.
        # Cheats with an +1 offset to avoid division by 0.
        zero_percentage = 100 / (len(line) / (line.count(0) + 1))

        if zero_percentage >= zero_threshold:
            return i

    # Only in case no mainly transparent line is found at all:
    return length


def panel_diameters(image):
    """Panel (or shadow ...) at the edge diameters.
    Wraps panel_diameter (used once per direction).

    """
    width = image.size[0]
    height = image.size[1]
    directions = ('rightwards', 'downwards', 'leftwards', 'upwards')
    lengths = (width, height, width, height)
    lefts_or_uppers = (0, 0, 0, 0)
    rights_or_lowers = (height, width, height, width)

    # map instead for-loop avoids predefining a list and dealing with
    # range/index. partial then avoids [image]*4 to make it iterable.
    return map(partial(panel_diameter, image),
               directions, lengths, lefts_or_uppers, rights_or_lowers)


def panel_filler(image, direction, length, left_or_upper,
                     right_or_lower):
    """One box that conforms to a part of a panel (or shadow ...) that can be
    repeated or cropped (and happens to be the longest (or last of the longest)
    part filled with the same pixel slices). Single pass, used 4 times via
    panel_boxes.

    """
    # Take a shortcut out, if the diameter is zero
    if left_or_upper == right_or_lower:
        start = 0
        stop = length

        return start, stop


    local_pixel_line = partial(pixel_line,
                               image,
                               moving_box(direction, length,
                                          left_or_upper, right_or_lower))

    max_repeats = 0
    repeats = 0
    stop = 0

    for i in range(length):
        line = local_pixel_line(i)

        # Compare lines. Track highest number of equal lines in succession.
        if i > 0:
            if previous_line == line:
                repeats += 1
                if repeats > max_repeats:
                    max_repeats = repeats
                    stop = i + 1
            else:
                repeats = 0

        previous_line = line

    start = stop - 1 - max_repeats

    return start,  stop


def panel_boxes(image):
    """Return boxes (left, upper, right, lower) corresponding to start and end
    parts of panels (or shadows, ...) and "empty" parts inbetween that can be
    repeated or cropped to fill more or less space.

    """
    # Analyze image
    # -------------

    width = image.size[0]
    height = image.size[1]

    # Get the alpha channel
    alpha = image.convert('RGBA').split()[-1]

    # Get diameters of panels (or shadows ...) on all 4 edges
    diameters = panel_diameters(alpha)

    # Start with left panel, downwards
    directions = ('downwards', 'rightwards', 'downwards', 'rightwards')
    lengths = (height, width, height, width)
    lefts_or_uppers = (0, 0, width - diameters[2], height - diameters[3])
    rights_or_lowers = (diameters[0], diameters[1], width, height)

    # map instead of for-loop avoids having to predefine a list and dealing
    # with range/index. partial then avoids [image]*4 to make it iterable.
    fillers = map(partial(panel_filler, image),
                      directions, lengths, lefts_or_uppers, rights_or_lowers)


    # Assemble a list of boxes
    # ------------------------
    # (left, upper, right, lower) per box. 3 boxes per edge plus one central

    # Parts per edge:
    # 'a' parts are to the left or above 'b' parts.
    # 'b' parts can be used to fill more or less space, as they consist of
    #  repeated 1 pixel wide slices
    # 'c' parts are to the right or below 'b' parts.

    # A simple listing would be more transparent, but making use of patterns is
    # more compact, more fun and might avoid certain types of errors :)
    # In the following, b like bottom stands for lowers, as l is for lefts.

    # List of lefts for left  edge,
    #         uppers  for upper edge,
    #         rights for right edge,
    #         lowers  for lower edge.
    # All of these are static for the a, b and c parts
    # and _don't_ depend on diameters.
    lurb = (0, 0, width, height)

    # List of uppers for left  edge,
    #         lefts  for upper edge,
    #         uppers for right edge,
    #         lefts  for lower edge
    ulul = [(0, fillers[i][0], fillers[i][1]) for i in range(4)]

    # List of rigths for left  edge,
    #         lowers for upper edge,
    #         lefts for right edge,
    #         uppers for lower edge.
    # All of these are static for the a, b and c parts
    # and _do_ depend on diameters.
    # abs() avoids negative values when subtracting from zero
    # in the first 2 rounds (afterwards from width and height)
    rblu = [abs(lurb[i] - diameters[i]) for i in range(4)]

    # List of lowers for left  edge,
    #         rights for upper edge,
    #         lowers for right edge,
    #         rights for lower edge.
    d = (height, width)*2
    brbr = [(fillers[i][0], fillers[i][1], d[i]) for i in range(4)]

    boxes = [[0]*4 for i in range(12)]
    for i in range(12):
        j = i // 3  # 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3
        k = (j + 1) % 2 # 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0
        l = i % 3       # 0, 1, 2
        n = (j + 2) % 4 # 2, 2, 2, 3, 3, 3, 0, 0, 0, 1, 1, 1
        p = k + 2       # 3, 3, 3, 2, 2, 2, 3, 3, 3, 2, 2, 2

        boxes[i][j] = lurb[j]
        boxes[i][k] = ulul[j][l]
        boxes[i][n] = rblu[j]
        boxes[i][p] = brbr[j][l]

    # Append box for the center part (content that doesn't belong to a panel)
    boxes.append((diameters[0], diameters[1],
                             width - diameters[2], height - diameters[3]))

    return boxes


def fill_regions(regions, target_rects):
    """Return images that use given images to fill the width and height
    specified in targets.

    """
    # Prepare list that will be returned
    filled = [0]*len(regions)

    for r in range(len(regions)):
        steps = []

        # Once for width, once for height
        for d in range(2):
            in_length = regions[r].size[d]
            out_length = target_rects[r][d]

            # How often does the region have to be repeated to cover the width
            # or height?
            times = int(ceil(out_length / max(in_length, 1)))


            # Create a list of positions to paste to for this dimension
            steps.append([in_length * n for n in range(times)])

        # Assemble coordinates list. Repeat lists to get a full lengh list out
        # of zip()
        coords = zip(steps[0]*len(steps[1]),
                     steps[1]*len(steps[0]))

        # Create blank image to past unto. target_rects values can be
        # negative, so make sure 0 is the lower limit.
        canvas = Image.new('RGBA', (max(target_rects[r][0], 0),
                                    max(target_rects[r][1], 0)))

        # Paste region unto canvas as often and where required
        for c in coords:
            canvas.paste(regions[r], c)

        # Update the dict that will be returned
        filled[r] = canvas

    return filled


def per_resolution(regions, (out_width, out_height), path):
    """Assemble and save one image. Called by output_images."""

    # Determine what goes where
    # -------------------------

    # Target positions to paste to
    # 10 items will remain 0
    # This instead of [[0]*2]*13] avoids references between list segments
    positions = [[0]*2 for i in range(13)]

    hw = (out_width, out_height)*2

    for i in range(1, 11, 3):
        j = i % 2
        # upper for left_b
        # left  for upper_b
        # upper for right_b
        # left  for lower_b
        positions[i][j] = regions[i - 1].size[j]

        # upper for left_c
        # left  for upper_c
        # upper for right_c
        # left  for lower_c
        positions[i + 1][j] = hw[j] - regions[i + 1].size[j]

    for i in range(6, 9):
        # lefts for right parts
        positions[i][0] = out_width - regions[6].size[0]

        # uppers for lower parts
        positions[i + 3][1] = out_height - regions[9].size[1]

    positions[12] = [regions[0].size[0], regions[3].size[1]]

    # Widths and heights that have to be filled
    targets = [[0]*2 for i in range(4)]
    for i in range(4):
        j = i % 2
        targets[i][j] = regions[i * 3].size[j]
        targets[i][j - 1] = hw[j - 1] - regions[i * 3].size[j] - regions[i * 3 + 2].size[j]

    # Create filler regions of the required sizes
    # -------------------------------------------

    fillers_pre = [regions[i] for i in range(1, 12, 3)]
    fillers = fill_regions(fillers_pre, targets)
    # Update regions list with results
    for i in range(4):
        j = i * 3 + 1
        regions[j] = fillers[i]


    # Assemble the image
    # ------------------

    # Combine all panel parts, leave out center
    # Overlapping areas in the corners are identical, so simple pasting is fine
    above = Image.new('RGBA', (out_width, out_height))
    for i in (range(len(regions) - 1)):
        # PIL doesn't seem to like lists, so use tuple()
        above.paste(regions[i], tuple(positions[i]))

    # Paste center region on its own full size canvas
    below = Image.new('RGBA', (out_width, out_height))
    below.paste(regions[12], tuple(positions[12]))

    # alpha_composite is slow, so only use it when and where required

    lower_upper = positions[9][1]
    center_left = positions[12][0]
    center_upper = positions[12][1]
    center_right = center_left + regions[12].size[0]
    right_left = positions[6][0]

    # Lower edge of center part, clipped to within image height
    center_lower = min(positions[12][1] + regions[12].size[1], out_height)

    comp_right = False
    comp_lower = False

    # Does the center extend to the right of the left edge of the right panel
    # and is there actually a panel?
    if right_left < center_right and right_left < out_width:
        overlap_right = (right_left, center_upper, out_width, center_lower)
        from_below = below.crop(overlap_right)
        from_above = above.crop(overlap_right)
        comp_right = alpha_composite(from_below, from_above)

    # Does the center extend to below the upper edge of the lower panel?
    # and is there actually a panel?
    if lower_upper < center_lower and lower_upper < out_height:
        overlap_bottom = (center_left, lower_upper, center_right, center_lower)
        from_below = below.crop(overlap_bottom)
        from_above = above.crop(overlap_bottom)
        comp_lower = alpha_composite(from_below, from_above)

    # Paste center region
    above.paste(regions[12], tuple(positions[12]))

    # If there's overlap, fix the areas that paste ruined
    if comp_right:
        above.paste(comp_right, overlap_right)
    if comp_lower:
        above.paste(comp_lower, overlap_bottom)


    # Save file
    # ---------

    # Save with filenames like 1024x0768.png
    # (zero padded on the left to 4 digits)
    above.save(''.join([path, "/",
                        str(out_width).zfill(4), "x",
                        str(out_height).zfill(4),
                        ".png"]))


def get_regions(args_image):
    image = Image.open(args_image)

    # Get the filename without extension
    file_name = common.filename_without_ext(args_image.name)

    # Get boxes (coordinates) for panel parts
    boxes = panel_boxes(image)
    # Copy parts from input image
    regions = map(image.crop, boxes)
    image_regions = [image]
    image_regions.extend(regions)

    return image_regions, file_name
