#!/usr/bin/env python

"""
Handle reading resolutions from a configuration file, writing that file if it
doesn't exist, or falling back to defaults. Used by bgc-sizes and sizes.


Copyright 2011 Thorsten Wilms
License: GPLv3, see docs/COPYING

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
import os.path
import sys


def read_or_create(path, default):
    """Read in resolutions from the configuration file specified by path.
    Create it, if it doesn't exist.

    """
    # Does the configuration file exist?
    if os.path.exists(path):
        # Configuration file exists, so read it

        f = open(path, 'r')
        l = list(f.read())
        f.close()

        # Filter out a list of numbers
        word = []
        resolutions = []
        pair = [0, 0]
        toggle = 0
        for i in l:
            try:
                # Collect numerals in word
                word.append(int(i))
            except:
                # Item wasn't a numeral and is a separator,
                # if it follows a numeral
                if len(word) > 0:
                    # previous item was a numeral, so it's the last part of a
                    # number. Alternate between assembling the first or second
                    # item of a pair
                    pair[toggle] = int(''.join([str(i) for i in word]))

                    if toggle == 0:
                        toggle = 1
                    else:
                        resolutions.append(tuple(pair))
                        toggle = 0
                # Reset, to collect a new word
                word = []

        if toggle == 1:
            text = "Warning: odd count of numbers in %s"
            print >> sys.stderr, text % (path)

    else:
        # Configuration file does not exist,
        # so write it and fall back to default

        f = open(path, 'w')
        f.write("List of output resolutions (width, height) for bgc-sizes and "
                "sizes:\n")
        for r in default:
            f.write(''.join([str(r[0]), " ", str(r[1]), "\n"]))
        f.close()

        resolutions = default

    return resolutions


def read_create_or_default():
    """Cause reading in resolutions from the configuration file specified by
    path. Cause creating it, if it doesn't exist. Fall back to default list of
    resolutions, if the home directory can't be found.

    """
    default = [(2560, 1600),
               (2560, 1440),
               (2048, 1152),
               (1920, 1200),
               (1920, 1080),
               (1680, 1050),
               (1680, 945),
               (1600, 1200),
               (1440, 900),
               (1366, 768),
               (1280, 1024),
               (1280, 800),
               (1024, 768),
               (1024, 576)]

    # Expand ~ to the user's home directory
    home = os.path.expanduser("~/")

    config = ".config/"
    conf_file = "backtestground_resolutions"

    if os.path.isdir(''.join([home, config])):
        resolutions = read_or_create(''.join([home, config,
                                                          conf_file]), default)
    elif os.path.isdir(home):
        resolutions = read_or_create(''.join([home, ".", conf_file]),
                                                 default)
    else:
        text = """Can't find your home directory to read
        %s (or to write it, if it doesn't exist).
        Falling back to a build in list of resolutions."""
        print text % (conf_file)

        resolutions = default

    return resolutions
