#!/usr/bin/env python

"""
Take an image and crop/resize it to several common screen resolutions.


Copyright 2010, 2011 Thorsten Wilms
License: GPLv3, see docs/COPYING

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
from __future__ import division
import Image

# Custom modules
import backtestground.common as common


def largest_fit(max_width, max_height, width, height):
    """Take 2 areas [width, height] and return the maximum area with the same
    aspect ratio as the 2nd pair, that fits within the first.

    """
    factor_width = max_width / width
    factor_height = max_height / height

    factor = min(factor_width, factor_height)

    result_width = int(round(width * factor))
    result_height = int(round(height * factor))

    return result_width, result_height


def per_resolution(image_list, (out_width, out_height), path):
    """Work to be done per resolution, up to saving the output image."""

    image = image_list[0]
    in_width = image.size[0]
    in_height = image.size[1]

    # Determine height and width of the part we need from the input image
    box_width, box_height = largest_fit(in_width, in_height,
                                        out_width, out_height)

    # Determine offsets needed due to the different aspect ratios
    offset_width = int(round((in_width - box_width) / 2))
    offset_height = int(round((in_height - box_height) / 2))

    # Copy appropriate area from the image
    # (left, upper, right, lower). Origin (0, 0) upper left corner.
    # Coordinates refer to positions between the pixels, so (0, 0, 0, 0)
    # doesn't work, while (0, 0, 1, 1) delivers 1 pixel.
    box = (offset_width, offset_height, offset_width + box_width,
           offset_height + box_height)
    region = image.crop(box)

    # Resize the region to final size
    out = region.resize((out_width, out_height), Image.BICUBIC)

    # Save with filenames like 1024x0768.png
    # (zero padded on the left to 4 digits)
    out.save(''.join([path, "/",
                        str(out_width).zfill(4), "x",
                        str(out_height).zfill(4),
                        ".png"]))


def get_regions(args_image):
    regions = [Image.open(args_image)]

    # Get the filename without extension
    file_name = common.filename_without_ext(args_image.name)

    return regions, file_name
